/***********************************************************************
 * File: sql_quoting.c
 * 	(Extracted from
 *	   "$Id: sql_quoting.c,v 1.2 2007/10/27 17:42:07 Administrator Exp Administrator $")
 *
 * Abstracts:
 *  This C source code implements the SQL escaping.
 *
 * Copyright (C) 2006-2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

/* include definitions for this API */
#include "sql_quoting.h"

/*
 * this API escapes a string for the SQL literal:
 *
 * arguments:
 *	1) src points to the source string.
 *	2) qualifier indicates the escaping methods.
 *	3) quote_char specifies quote code of `src' literal in SQL.
 *	4) escape_code specifies a code in SQL `ESCAPE' statement.
 *	5) dst points to the destiny buffer. this API outputs escaped
 *	  string toward it. (size of the buffer should be
 *	  "strlen(src) * 2 + 3")
 *	6) dst_sz indicates the size of destiny buffer.
 *
 * returns:
 *	Not NULL (same as dst) means normal end.
 *	NULL occurred several errors. errno indicates the detail:
 *		EINVAL:	  illegal arguments
 *		EMSGSIZE: buffer over run (size of the destiny buffer is
 *			  insufficient)
 */
char *sql_quoting(char *src
		, SQL_quoting_qualifier_t qualifier
		, char quote_char
		, char escape_code
		, char *dst
		, size_t dst_sz
		)
{
	char	quote_buf[2], escape_buf[2], *result = dst;

	/* test: illegal arguments? */
	if ((((((src == NULL)
	|| dst == NULL)
	|| strlen(src) >= dst_sz)
	|| !is_legal_SQL_quoting_qualifier(qualifier))
	|| !is_legal_SQL_quote_character(quote_char))
	|| !is_legal_SQL_escape_code(escape_code, qualifier)) {
		errno	= EINVAL; return NULL; /* illegal arguments */
	}

	/* append a character to dst (checking buffer over run) { */
#	define APPEND_CHR(DST,SRC,DST_SZ) {			\
		if (1 >= (DST_SZ)) {				\
			/* buffer overflow */			\
			errno = EMSGSIZE; return NULL;		\
		}						\
		strcat((DST),(SRC)); ++(DST); --DST_SZ;		\
	}

	*dst	= '\0'; /* make dst an empty string */

	/* make a string to hold quote character */
	quote_buf[0] = quote_char; quote_buf[1] = '\0';

	/* make a string to hold escape code */
	if ((SQL_quoting_escape_pattern_codes & qualifier) != 0) {
		escape_buf[0]	= (escape_code == '\0')?
						'\\': escape_code;
		escape_buf[1]	= '\0';
	}
	else {
		escape_buf[0]	= '\0';
	}

	/* enclose dst with specified quote characters */
	if ((SQL_quoting_enclose_result & qualifier) != 0) {
		APPEND_CHR(dst, quote_buf, dst_sz);
	}

	/* process src... */
	while ((dst_sz > 1) && *src != '\0') {
		/* test: find the specified quote character? */
		if (((SQL_quoting_do_not_escape_quots & qualifier)
								!= 0)
		&& *src == quote_char) {
			/* escape the character with a backslash or
			 * its self (default is its self)
			 */
			APPEND_CHR(dst
				, (((qualifier
					& SQL_quoting_with_backslash)
					!= 0)?	((escape_code == '\0')?
							  "\\"
							: escape_buf
						)
						: quote_buf
					)
				, dst_sz
				);
		}

		/* test: find a control character (specified escape it),
		 *	backslash, percent-sign, under score, vertical
		 *	bar?
		 */
		if (((((((SQL_quoting_escape_control_codes & qualifier)
								!= 0)
			&& strchr(SQL_quoting_CONTROL_CHARACTERS, *src)
							!= NULL)
			)
		|| *src == escape_code)
		|| (((SQL_quoting_escape_vertical_bar & qualifier) != 0)
			&& *src == '|'))
		|| (((SQL_quoting_escape_pattern_codes & qualifier)
								!= 0)
			&& (*src == '%' || *src == '_'))
		) {
			if (escape_buf[0] != '\0') {
				/* escape it with a backslash sign */
				APPEND_CHR(dst, escape_buf, dst_sz);
			}
		}

		/* test: dst has no space? */
		if (dst_sz <= 1) {
			/* buffer overflow */
			errno	= EMSGSIZE; return NULL;
		}

		/* append the src character to the dst buffer */
		*dst++ = *src++; --dst_sz; *dst = '\0';
	}

	/* test: src has rest? */
	if (*src != '\0') {
		errno = EMSGSIZE; return NULL; /* buffer overflow */
	}

	/* enclose dst with specified quote characters */
	if ((SQL_quoting_enclose_result & qualifier) != 0) {
		APPEND_CHR(dst, quote_buf, dst_sz);
	}

	return	result; /* normal end: address to destiny buffer */
#	undef APPEND_CHR /* } */
}
